__version__ = "0.0.8"
__authors__ = ["Jacob Jeffries", "Ameen Mahmood"]
__author_emails__ = ["jacob.jeffries@chicagoclubteam.org", "ameen.mahmood@chicagoclubteam.org"]
__url__ = "https://github.com/Chicago-Club-Management-Company/aqua-blue"

from .reservoir import EchoStateNetwork as EchoStateNetwork
from .time_series import TimeSeries as TimeSeries
